#include<string.h>

#include "softAllDiff.h"
#include "hopcroft-karp.h"

void DFS(int sizeOfU, int sizeOfV, int u, int* color, int* nbSucc, int succ[sizeOfV][sizeOfU], int* matchedWithU, int* order, int* nb){
  // perform a depth first search, starting from u, in the bipartite graph Go=(U,V,E) such that
  // U = vertices of Gp
  // V = vertices of Gt
  // E = { (u,matchedWithU[u]) / u is a vertex of Gp } U
  //     { (v,u) / v is a vertex of D[u] which is not matched to v}
  // Given a vertex v of Gt, nbSucc[v]=number of sucessors of v and succ[v]=list of successors of v
  // order[nb^out+1..nb^in] contains the vertices discovered by the DFS
  int i;
  color[u]=grey;
  int v=matchedWithU[u]; // the only one predecessor of v is u
  if(v>=0)
  {
  for (i=0; i<nbSucc[v]; i++)
    if (color[succ[v][i]]==white) 
      DFS(sizeOfU,sizeOfV,succ[v][i],color,nbSucc,succ,matchedWithU,order,nb);}
  // we have finished with u => number it
  order[*nb]=u; (*nb)--;
}

void SCC(int sizeOfU, int sizeOfV, int* markedV, int* markedU, 
	 int* nbSucc, int succ[sizeOfV][sizeOfU], 
	 int* nbPred, int pred[sizeOfU][sizeOfV], int* matchedWithU,int* matchedWithV){
  // postrelation: markedV[v]==markedU[u] iff they belong to the same
  // strongly connected component in the bipartite graph Go=(U,V,E) such that
  // U = vertices of Gp
  // V = vertices of Gt
  // E = { (u,matchedWithU[u]) / u is a vertex of Gp } U
  //     { (v,u) / v is a vertex of D[u] which is not matched to v}
  // Given a vertex v of Gt, nbSucc[v]=number of sucessors of v and succ[v]=list of successors of v
  int order[sizeOfU];
  int color[sizeOfU];
  int fifo[sizeOfV];
  int u, v, i, j, k, nbSCC, nb;
	
  // Order vertices of Gp wrt DFS
  memset(color,white,sizeOfU*sizeof(int));
  nb=sizeOfU-1;
  for (u=0; u<sizeOfU; u++)
    if (color[u]==white) DFS(sizeOfU,sizeOfV,u,color,nbSucc,succ,matchedWithU,order,&nb);
	
  // traversal starting from order[0], then order[1], ...
  nbSCC=0;
  memset(markedU,-1,sizeOfU*sizeof(int));
  memset(markedV,-1,sizeOfV*sizeof(int));
  for (i=0; i<sizeOfU; i++){
    u=order[i];
    v=matchedWithU[u];
    if(v>=0){
    if (markedV[v]==-1){ // v belongs to a new SCC
      nbSCC++;
      k=1; fifo[0]=v;
      markedV[v]=nbSCC;
      while (k>0){
	v=fifo[--k];
	u=matchedWithV[v];
	if (u!=-1){
	  markedU[u]=nbSCC;
	  for (j=0; j<nbPred[u]; j++){
	    v=pred[u][j];
	    if (markedV[v]==-1){
	      markedV[v]=nbSCC;
	      fifo[k++]=v;
	    }
	  }
	}
      }
    }
   }
  }
}


int checkSoftAllDiff(
		 int sizeOfU, // number of elements in U
		 int sizeOfV, // number of elements in V
		 Tdomain* D, // for each element u of U, D represents the elements of V compatible with u at a given depth in the search tree
		 int prof, // depth in the search tree 
		 int* matchedWithU, // matchedWithU[i] = element of V matched with the ith element of U
		 int* nbToRemove, // (*nbToRemove) = number of values to remove from domains
		 int* toRemove, // remove toRemove[2i+1] from the domain of toRemove[2i]
		 int maxSize // best current solution : maxSize+1 elements in U must be matched to improve the current best solution
		 ){
  // compute a largest matching matchedWithU
  // compute toRemove elements of V that does not appear in any largest matching, if the size of the matching is equal to maxSize+1, 
  // return the size of the matching
  
  int nbMatched;
	
  nbMatched=updateMatching(sizeOfU,sizeOfV,D,prof,matchedWithU);
  
  if(nbMatched==maxSize+1)
  {

  // Filter domains to remove values which cannot belong to any largest matching
  // Build the bipartite directed graph Go=(U,V,E) such that
  // E = { (u,v) / u is a vertex of U which is matched to v (i.e., v=matchedWithU[u])} U
  //     { (v,u) / v is a vertex of V which is compatible with u, but is not matched to u}
  int nbPred[sizeOfU]; // nbPred[u] = nb of predecessors of u in Go
  int pred[sizeOfU][sizeOfV]; // pred[u][i] = ith predecessor of u in Go
  int nbSucc[sizeOfV]; // nbSucc[v] = nb of successors of v in Go
  int succ[sizeOfV][sizeOfU]; // succ[v][i] = ith successor of v in Go
  int u,v,i,w;
  int markedV[sizeOfV], markedU[sizeOfU];
  int used[sizeOfU][sizeOfV];
  int matchedWithV[sizeOfV];
  memset(markedV,white,sizeOfV*sizeof(int));
  memset(nbSucc,0,sizeOfV*sizeof(int));
  memset(matchedWithV,-1,sizeOfV*sizeof(int));
  memset(nbPred,0,sizeOfU*sizeof(int));
  memset(markedU,white,sizeOfU*sizeof(int));
  
  for (u=0; u<sizeOfU; u++) for (v=0; v<sizeOfV; v++) used[u][v]=0;
  
  for (u=0; u<sizeOfU; u++){
    if(matchedWithU[u] >=0) matchedWithV[matchedWithU[u]] = u;

    if(D->current_sol[u]==-1)
    {
      	for (i=D->firstVal[u]; i<D->firstVal[u]+D->nbVal_prof[prof][u]; i++)
      	{
        	v=D->val[i];
      		if (v != matchedWithU[u])
      		{
			pred[u][nbPred[u]++]=v; 
			succ[v][nbSucc[v]++]=u; 
      		}
    	}
    }
  }
  
	
  // mark as used all edges of paths starting from free vertices
  int list[sizeOfV];
  int nb=0;
  for (v=0; v<sizeOfV; v++){
    if (matchedWithV[v] < 0){ // v is free
      list[nb++]=v;
      markedV[v]=grey;
    }
  }
  while (nb>0){
    v=list[--nb];
    for (i=0; i<nbSucc[v]; i++){
      u=succ[v][i];
      used[u][v]=1;
      if (markedU[u]==white){
	markedU[u]=grey;
	w=matchedWithU[u];
	used[u][w]=1;
	if (markedV[w]==white){
	  list[nb++]=w;
	  markedV[w]=grey;
	}
      }
    }
  }
 
 
 
  // look for strongly connected components in Go

  SCC(sizeOfU,sizeOfV,markedV,markedU,nbSucc,succ,nbPred,pred,matchedWithU,matchedWithV);
	
  // remove v from D[u] if (u,v) is not marked as used 
  //                    and u and v are not in the same SCC 
  //                    and matchedWithU[u] != v
  (*nbToRemove) = 0;
  for (u=0; u<sizeOfU; u++){
    for (i=D->firstVal[u]; i<D->firstVal[u]+D->nbVal_prof[prof][u]; i++){
      v=D->val[i];
      if ((used[u][v]==0) && (markedV[v]!=markedU[u]) && (matchedWithU[u]!=v)){
	toRemove[2*(*nbToRemove)] = u;
	toRemove[2*(*nbToRemove)+1] = v;
	(*nbToRemove)++;
      }
    }
  }
 }
  return nbMatched;
}


int checkSoftAllDiff_lazy(
		 int sizeOfU, // number of elements in U
		 int sizeOfV, // number of elements in V
		 Tdomain* D, // for each element u of U, D represents the elements of V compatible with u at a given depth in the search tree
		 int prof, // depth in the search tree 
		 int* matchedWithU, // matchedWithU[i] = element of V matched with the ith element of U
		 int* nbToRemove, // (*nbToRemove) = number of values to remove from domains
		 int* toRemove, // remove toRemove[2i+1] from the domain of toRemove[2i]
		 int maxSize // best current solution : maxSize+1 elements in U must be matched to improve the current solution 
		 ){
  // try to compute a matching, matchedWithU, whose size is greater than maxSize+1 (and stops as soon as it is achieved) if there exists any  
  // compute in toRemove elements of V that does not appear in any largest matching, if the size of the matching is equal to maxSize+1 
  // return the size of the matching
  
  int nbMatched;
	
  nbMatched=updateMatching_lazy(sizeOfU,sizeOfV,D,prof,matchedWithU,maxSize);
 
  if(nbMatched==maxSize+1)
  {
   
  // Filter domains to remove values which cannot belong to any largest matching
  // Build the bipartite directed graph Go=(U,V,E) such that
  // E = { (u,v) / u is a vertex of U which is matched to v (i.e., v=matchedWithU[u])} U
  //     { (v,u) / v is a vertex of V which is compatible with u, but is not matched to u}
  int nbPred[sizeOfU]; // nbPred[u] = nb of predecessors of u in Go
  int pred[sizeOfU][sizeOfV]; // pred[u][i] = ith predecessor of u in Go
  int nbSucc[sizeOfV]; // nbSucc[v] = nb of successors of v in Go
  int succ[sizeOfV][sizeOfU]; // succ[v][i] = ith successor of v in Go
  int u,v,i,w;
  int markedV[sizeOfV], markedU[sizeOfU];
  int used[sizeOfU][sizeOfV];
  int matchedWithV[sizeOfV];
  memset(markedV,white,sizeOfV*sizeof(int));
  memset(nbSucc,0,sizeOfV*sizeof(int));
  memset(matchedWithV,-1,sizeOfV*sizeof(int));
  memset(nbPred,0,sizeOfU*sizeof(int));
  memset(markedU,white,sizeOfU*sizeof(int));
  
  for (u=0; u<sizeOfU; u++) for (v=0; v<sizeOfV; v++) used[u][v]=0;
  
  for (u=0; u<sizeOfU; u++){
    if(matchedWithU[u] >=0) matchedWithV[matchedWithU[u]] = u;

    if(D->current_sol[u]==-1)
    {
      	for (i=D->firstVal[u]; i<D->firstVal[u]+D->nbVal_prof[prof][u]; i++)
      	{
        	v=D->val[i];
      		if (v != matchedWithU[u])
      		{
			pred[u][nbPred[u]++]=v; 
			succ[v][nbSucc[v]++]=u; 
      		}
    	}
    }
  }
  
	
  // mark as used all edges of paths starting from free vertices
  int list[sizeOfV];
  int nb=0;
  for (v=0; v<sizeOfV; v++){
    if (matchedWithV[v] < 0){ // v is free
      list[nb++]=v;
      markedV[v]=grey;
    }
  }
  while (nb>0){
    v=list[--nb];
    for (i=0; i<nbSucc[v]; i++){
      u=succ[v][i];
      used[u][v]=1;
      if (markedU[u]==white){
	markedU[u]=grey;
	w=matchedWithU[u];
	used[u][w]=1;
	if (markedV[w]==white){
	  list[nb++]=w;
	  markedV[w]=grey;
	}
      }
    }
  }
 
 	
  // look for strongly connected components in Go

  SCC(sizeOfU,sizeOfV,markedV,markedU,nbSucc,succ,nbPred,pred,matchedWithU,matchedWithV);
	
  // remove v from D[u] if (u,v) is not marked as used 
  //                    and u and v are not in the same SCC 
  //                    and matchedWithU[u] != v
  (*nbToRemove) = 0;
  for (u=0; u<sizeOfU; u++){
    for (i=D->firstVal[u]; i<D->firstVal[u]+D->nbVal_prof[prof][u]; i++){
      v=D->val[i];
      if ((used[u][v]==0) && (markedV[v]!=markedU[u]) && (matchedWithU[u]!=v)){
	toRemove[2*(*nbToRemove)] = u;
	toRemove[2*(*nbToRemove)+1] = v;
	(*nbToRemove)++;
      }
    }
  }
 }
 
  return nbMatched;
}

